open Format
open Options

module Rat = struct
  open Num
  type t = num

  let of_int = num_of_string
  let zero = Int 0
  let one = Int 1
  let minusone = Int (-1)
  let plus = add_num
  let minus = sub_num
  let mult = mult_num
  let inverse x = div_num minusone x
  let abs = abs_num
  let compare = compare_num
  let print fmt x = fprintf fmt "%s" (string_of_num x)
end

module P = Polynomes.Make(Rat)

module R = struct
  type elt = P.t

  module S = Term.Set
  module M = Term.Map
  module A = Literal
  module SA = A.Set
  module SAEq = A.SetEq
  module SP = P.Set

  let ale = Hstring.make "<=" 
  let alt = Hstring.make "<"
    
  type ineq = { i: P.t ; d: SA.t ; k: bool (*true: <=*) }
      
  module SPA ( E : sig type g end) = struct
    include Set.Make
      (struct 
	 type t = A.t * E.g
	 let compare (a1,_) (a2,_) = A.compare a1 a2
       end)
    let extract s =  fold (fun (_,e) inqs -> e::inqs) s []
  end

  module SPAi = SPA(struct type g=ineq end)
  module SPAii = SPA(struct type g=ineq * ineq end)

  type t = { 
    inqs : SPAi.t;
    monomes : S.t ; 
    tighten : SPAii.t }
      
  let empty = { 
    inqs = SPAi.empty ; 
    monomes = S.empty ; 
    tighten = SPAii.empty  }

  exception Inconsistent
  exception NotConsistent of SA.t

  let add_monomes env p = 
    {env with monomes = M.fold (fun t _ -> S.add t) (P.dumview p) env.monomes }
          
  let checking_monomes env (_,(ip1,ip2)) =
    let s = S.union (P.leaves ip1.i) (P.leaves ip2.i) in
    try
      S.iter (fun t ->if S.mem t env.monomes then raise Exit) s; false
    with Exit -> true

  let reconsidering env all_done = 
    SPAii.diff (SPAii.filter (checking_monomes env) env.tighten) all_done

  let is_le n = Hstring.compare n ale = 0
  let is_lt n = Hstring.compare n alt = 0

  let zero = Rat.zero
  let mun = Rat.minusone
  let p_un = P.make (Term.int "1")
  let p_zero = P.make (Term.int "0")

  let remove_sa env sa = 
    let i = { i=p_zero; d=SA.empty ; k=true} in
    SA.fold 
    (fun a env -> {env with inqs = SPAi.remove (a,i) env.inqs }) sa env

  let is_simple m k d = 
    let (m,a) = P.view m in
    M.is_empty m && ( 
      (k && Rat.compare a zero <= 0) || 
	(not k && Rat.compare a zero <0 ) || raise (NotConsistent d))
      
  let is_eq m k = 
    let (m,a) = P.view m in k && M.is_empty m && Rat.compare a zero = 0

  let cross x cpos = 
    let cpos = 
      List.map (fun ({i=i} as e) -> M.find x (P.dumview i) , e) cpos in
    let rec cross_rec r = function 
      |	[] -> r
      | {i=i1;d=d1;k=k1} :: l -> 
	  let n1 = M.find x (P.dumview i1) in
	  let r = 
	    List.fold_left 
	      (fun r (n2,{i=i2;d=d2;k=k2}) ->
		 let ni = 
		   { i=P.add (P.mult (Rat.abs n2) i1) (P.mult (Rat.abs n1) i2);
		     d=SA.union d1 d2;
		     k=k1&&k2}
		 in ni::r) r cpos
	  in cross_rec r l
    in cross_rec []

  let split x = 
    let rec split_rec ((cp,cn,co) as r) = function
	[] -> r
      | ({i=p} as e)::l -> 
	  let m, _ = P.view p in
	  let r = (
	    try 
	      if Rat.compare (M.find x m) zero > 0 then e::cp , cn , co 
	      else cp , e::cn , co
	    with Not_found -> cp , cn , e::co)
	  in split_rec r l
    in split_rec ([],[],[])
   
  let fm = 
    let rec fourier r l = match l with
	[] -> r
      | {i=i;d=d;k=k} :: l' when is_simple i k d ->
	  fourier (if is_eq i k then SA.union d r else r) l'
      | {i=i;d=d;k=k} :: _  ->
	  let x,_ = P.choose i in
	  let cpos , cneg , others = split x l in
	  let ninqs = cross x cpos cneg in
	  fourier r (ninqs@others)
    in fourier SA.empty 
	 
  let assume_le_lt env leqs = 
    let env = 
      List.fold_left
	(fun env ((a,_) as ai) ->
	   if debug_fm then eprintf "[fm] we assume %a@." A.print a;
	   {env with inqs=SPAi.add ai (SPAi.remove ai env.inqs) }
	) env leqs
    in
    env , fm (SPAi.extract env.inqs)
      
  let backtrack_neq env sa a ip1 ip2 =
    let env'  = { env with tighten=SPAii.remove (a,(ip1,ip2)) env.tighten } in
    let env1 = 
      add_monomes { env' with inqs = SPAi.add (a,ip1) env'.inqs } ip1.i in
    let env2 = 
      add_monomes { env' with inqs = SPAi.add (a,ip2) env'.inqs } ip2.i in
    try 
      let r = fm (SPAi.extract env1.inqs) in
      (try
	 let _ = fm (SPAi.extract env2.inqs) in 
	 if debug_fm then Format.eprintf "[fm] We delay %a@." A.print a;
	 env ,  sa
       with NotConsistent _ -> env1 , SA.union r sa)
    with NotConsistent _ -> 
      env2 , SA.union sa (fm (SPAi.extract env2.inqs))
      
  let assume_neqs = 
    let rec assume_rec all_done env sa neqs = 
      let todo , delay = SPAii.partition (checking_monomes env) neqs in
      let rc = reconsidering env all_done in
      let todo = SPAii.union rc todo in
      if SPAii.is_empty todo then env , sa    
      else 
	let env , sa = 
	  SPAii.fold
	    (fun (a,(ip1,ip2)) (env,sa) ->
	       if debug_fm then 
		 Format.eprintf "[fm] We try %a@." A.print a;
	       backtrack_neq env sa a ip1 ip2) todo (env,sa)
	in
	assume_rec (SPAii.union rc all_done) env sa delay
    in assume_rec SPAii.empty

  let new_equalities sa = 
    SA.fold (fun a saeq -> match A.view a with
		 A.Built(_,n,[t1;t2]) when is_le n -> SAEq.add (a,t1,t2) saeq
	       | _ -> saeq) sa SAEq.empty

  let normal_form a = 
    match A.view a with
	A.Built(false,n,[t1;t2]) when is_le n && Term.is_int t1 ->
	  A.make (A.Built(true,n,[t2;Term.pred t1]))
      | A.Built(false,n,[t1;t2]) when is_le n -> 
	  A.make (A.Built(true,alt,[t2;t1]))
      | A.Built(false,n,[t1;t2]) when is_lt n ->
	  A.make (A.Built(true,ale,[t2;t1]))
      | _ -> a

  let assume env sa repr _ = try
    let env , neqs , ineqs = 
      SA.fold
	(fun a (env , neqs , ineqs) ->
	   match A.view a with
	     | A.Built(_,n,[t1;t2]) when is_le n || is_lt n ->
		 let p1 = 
		   P.add (repr t1) (P.mult mun (repr t2)) in
		 let ai = a , {i=p1;d=SA.singleton a;k=is_le n} in
		 add_monomes env p1 , neqs , ai::ineqs
	     | A.Neq(t1,t2) when Term.is_int t1 && Term.is_int t2 -> 
		 let r1 , r2 = repr t1 , repr t2 in
		 let p1 = P.add r1 (P.mult mun r2) in
		 let p2 = P.add r2 (P.mult mun r1) in
		 let p1 , p2 = 
		   if not(Term.is_int t1) then p1,p2
		   else P.add p_un p1 , P.add p_un p2 in
		 let ip1 = { i=p1; d=SA.singleton a; k=true } in
		 let ip2 = { i=p2; d=SA.singleton a; k=true } in
		 let env = 
		   { env with tighten = SPAii.add (a,(ip1,ip2)) env.tighten }
		 in
		 env , SPAii.add (a,(ip1,ip2)) neqs , ineqs
	     | _ -> (env , neqs , ineqs) ) sa (env,SPAii.empty,[])
    in
    let env , sa = assume_le_lt env ineqs in
    let env , sa = assume_neqs env sa neqs in
    remove_sa env sa , (new_equalities sa)
  with NotConsistent _ ->  raise Inconsistent

  let query a repr cl env = 
    try 
      let a = normal_form a in
      ignore(assume env (SA.singleton a) repr cl); 
      false 
    with Inconsistent -> true
      
end

include P

let is_mine_t t =  
(*  (Term.view t).Term.ty = Ty.Tint*)
  match  Term.view t with
    | {Term.f=Symbols.Int _} | {Term.f=Symbols.Binop _} -> true
    | {Term.f=Symbols.Name _ ; xs=[] ; ty = Ty.Tint} -> true
    | _ -> false
      
let is_mine_r = is_var

let is_mine_a a = 
  match Literal.view a with
    | Literal.Built (_,p,_) -> R.is_le p || R.is_lt p
    | _ -> false

let normal_form = R.normal_form

let name = "arith "
