open Format
open Hashcons
open Options

module T = Term
module Sy = Symbols

type lemma = {
  vars: Sy.Set.t;
  trs : T.t list list;
  f : t;
  name : string
}
    
and view = 
    U of t list
  | C of t*t  
  | Lit of Literal.t
  | Lem of lemma 
  | Sko of T.t Subst.t * t

and iview = { pos : view ; neg : view }

and t = iview hash_consed
    
module View = struct
  type t = iview
      
  let rec compare_list compare l1 l2 = match l1 , l2 with
      [] , [] -> 0
    | [] , _ -> 1
    | _ , [] -> -1
    | x1::l1 , x2::l2 -> 
	let c = compare x1 x2 in if c<>0 
	then c else compare_list compare l1 l2
	
  let rec compare_pclause v1 v2 = match v1 , v2 with
      U l1 , U l2 -> compare_list compare_t l1 l2
    | U _ , _ -> -1
    | _, U _ -> 1
    | C(x1,y1) , C(x2,y2) -> 
	  let c = compare_t x1 x2 in if c<>0 then c else compare_t y1 y2
    | C _ , _ -> -1
    | _ , C _ -> 1
    | Lit a1 , Lit a2 ->Literal.compare a1 a2
    | Lit _ , _ -> -1
    | _ , Lit _ -> 1
    | Lem l1 , Lem l2 -> compare_lemme l1 l2
    | Lem _ , _ -> -1
    | _ , Lem _ -> 1
    | Sko(s1,f1) , Sko(s2,f2) -> compare_skolem s1 f1 s2 f2
	
  and compare_t t1 t2  = Pervasives.compare t1.tag t2.tag
  and compare_view v1 v2 = 
    let c = compare_pclause v1.pos v2.pos in
    if c<>0 then c else compare_pclause v1.neg v2.neg
  and compare_lemme l1 l2 = 
    let c = compare_t l1.f l2.f in
    if c<>0 then c else
      let c = Sy.Set.compare l1.vars l2.vars in
      if c<>0 then c else 
	compare_list (compare_list T.compare) l1.trs l2.trs
  and compare_skolem s1 f1 s2 f2 = 
    let c = compare_t f1 f2 in
    if c<>0 then c else Sy.Map.compare T.compare s1 s2
      
  let sort l = 
    let l = List.sort compare_t l in
    List.fold_left 
	(fun acc x -> match acc with
	     [] -> [x]
	   | h::l when x==h -> acc
	   | _ -> x::acc) [] l
      
  let eqc c1 c2 = match c1,c2 with
      U l1 , U l2 -> 
	(try 
	   List.for_all2 (==) (sort l1) (sort l2)
	 with Invalid_argument _ -> false)
    | C(f1,f2) , C(g1,g2) -> f1==g1 && f2==g2 || f1==g2 && f2==g1
    | Lit x , Lit y -> Literal.equal x y
    | Lem({trs=lt1;f=f1}) ,Lem({trs=lt2;f=f2}) -> 
	(try List.for_all2 (List.for_all2 T.equal) lt1 lt2 && f1==f2
	 with Invalid_argument _ -> false)
    | Sko(s1,f1) , Sko(s2,f2) -> 
	f1==f2 && Sy.Map.compare T.compare s1 s2 = 0
    | _, _ -> false
	
  let hashlt = List.fold_left (fun acc x->acc*19 + T.hash x)
  let hashllt = List.fold_left (fun acc x->acc*19 + hashlt 0 x)
    
  let hashc acc = function 
      U l -> List.fold_left (fun acc f -> acc * 19 + f.tag) 1 (sort l)
    | C(f1,f2) -> 
	let min = min f1.tag f2.tag in
	let max = max f1.tag f2.tag in
	(acc*19 + min)*19 + max
    | Lem {vars=vars;trs=trs;f=f} -> 
	hashllt (Hashtbl.hash (f.tag,vars)) trs
    | Lit x -> Literal.hash x
    | Sko(s,f) -> 
	Sy.Map.fold 
	  (fun s t acc ->acc * 19 + Sy.hash s) s f.tag
	  
  let equal f1 f2 = eqc f1.pos f2.pos && eqc f1.neg f2.neg
  let hash f = abs (hashc (hashc 1 f.pos) f.neg)
end
  
let compare f1 f2 = compare f1.tag f2.tag
let equal f1 f2 = f1.tag == f2.tag
  
module H = Make(View)
  
let tbl = H.create 100007
let iview f = f.node

(* this function should only be applied with ground substitutions *)
let rec apply_subst subst f = 
  let {pos=p;neg=n} = iview f in
  H.hashcons tbl {pos=iapply_subst subst p; neg=iapply_subst subst n}

and iapply_subst ((s_t,s_ty) as subst) = function
    Lit a -> 
      Lit(Literal.apply_subst subst a)
  | Lem({vars=vars; trs=trs; f=f} as e)->
      let s_t = Sy.Set.fold Sy.Map.remove vars s_t in
      let subst = s_t , s_ty in
      let f =  apply_subst subst f in
      let trs = List.map (List.map (T.apply_subst subst)) trs in
      Lem({e with trs=trs; f=f})
  | U l -> 
      U(List.map (apply_subst subst) l)
  | C(f1,f2) -> 
      C(apply_subst subst f1,apply_subst subst f2)
  | Sko(s_t',f) -> 
      Sko(s_t',apply_subst subst (apply_subst (s_t',s_ty) f))
	(* a refaire car danger d'explosion et de clash !*)
      
    
let view t = t.node.pos

let rec print fmt f = match view f with
    Lit a -> Literal.print fmt a
  | Lem {trs=trs;f=f;name=n} -> 
      if verbose then
	fprintf fmt "[%a] %a" 
	  (fun fmt -> List.iter (fprintf fmt "%a |" T.print_list)) 
	  trs print f
      else 
	fprintf fmt "%s" n
(*      fprintf fmt "<lem> %a" print f*)
  | U l -> List.iter (fun f->fprintf fmt "%a & " print f) l
  | C(f1,f2) -> fprintf fmt "(%a v %a) " print f1 print f2
  | Sko(s,f) -> fprintf fmt "<sko> (%a)" print f
      
    
(* smart constructors *)

let mk_lit a = H.hashcons tbl {pos=Lit a; neg=Lit(Literal.neg a)}
  
let mk_not f = 
  let f = iview f in H.hashcons tbl ({pos=f.neg;neg=f.pos})	  

let mk_skolem_subst bv v = 
  T.Set.fold 
    (fun x m -> 
       let {T.f=x;ty=ty} = T.view x in
       let bv = T.Set.fold (fun y acc-> y::acc) bv [] in
       let t = T.make (Sy.fresh ()) bv ty in
       Sy.Map.add x t m) 
    v Sy.Map.empty

let symbols_of_terms v = 
  T.Set.fold 
    (fun t sy -> let {T.f=f} = T.view t in Sy.Set.add f sy) 
    v Sy.Set.empty
  
(* name: (forall bv [trs]. f[fv]) *)
let mk_forall fv up bv trs f name = 
  let sy = symbols_of_terms bv in
  let lem = { vars=sy; trs=trs; f=f ; name=name} in
  let subst = mk_skolem_subst up bv in
  H.hashcons tbl {pos=Lem(lem) ; neg=Sko(subst,mk_not f) }
    
(* forall upbv.  name: (exists bv [trs]. f) *)
let mk_exists fv up bv trs f name = 
  let subst = mk_skolem_subst up bv in
  let sy = symbols_of_terms bv in
  let lem = {vars=sy; trs=trs; f=mk_not f; name=name} in
  H.hashcons tbl {pos=Sko(subst,f)  ; neg=Lem(lem)}
    
let mk_and f1 f2 = 
  if equal f1 f2 then f1 else
    H.hashcons tbl {pos=U([f1;f2]); neg=C(mk_not f1,mk_not f2)}
      
let mk_or f1 f2 = 
  if equal f1 f2 then f1 else
    H.hashcons tbl {pos=C(f1,f2); neg=U([mk_not f1;mk_not f2])}
      
let mk_imp f1 f2 = 
  H.hashcons tbl {pos=C(mk_not f1,f2); neg=U([f1;mk_not f2])}

let mk_if f1 f2 f3 = 
  mk_or (mk_and f1 f2) (mk_and (mk_not f1) f3)
    
let mk_iff f1 f2 = 
  let a = mk_or f1 f2 in
  let b = mk_or (mk_not f1) (mk_not f2) in
  let c = mk_or (mk_not f1) f2 in
  let d = mk_or f1 (mk_not f2) in
  H.hashcons tbl {pos=U([c;d]); neg=U([a;b])}

let free_vars =
  let rec free_rec acc f = 
    match view f with
	Lit a -> Sy.Set.union (Literal.vars_of a) acc
      | Lem {vars=v;f=f} -> let s = free_rec acc f in Sy.Set.diff s v
      | U l -> List.fold_left free_rec acc l
      | C(f1,f2) -> free_rec (free_rec acc f1) f2
      | Sko(subst,f) -> 
	  let sy = free_rec acc f in
	  Sy.Map.fold 
	    (fun s t sy -> 
	       if Sy.Set.mem s sy then 
		 Sy.Set.remove s (Sy.Set.union sy (Term.vars_of t))
	       else sy
	    ) subst sy
  in free_rec Sy.Set.empty
    
let terms = 
  let rec terms acc f = match view f with
      Lit a -> 
	let s = 
	  T.Set.filter 
	    (fun t->Sy.Set.is_empty (T.vars_of t)) (Literal.terms_of a)
	in
	T.Set.union s acc
    | Lem {trs=trs;f=f} -> terms acc f
    | U l -> List.fold_left terms acc l
    | C(f1,f2) -> terms (terms acc f1) f2
    | Sko(s,f) -> terms acc f
  in terms T.Set.empty

    

module Set = Set.Make(struct type t'=t type t=t' let compare=compare end)
module Map = Map.Make(struct type t'=t type t=t' let compare=compare end)
