open Format
open Options

let name = "pairs "

type abstract = 
    Pair of abstract * abstract 
  | Fst of abstract 
  | Snd of abstract 
  | Other of Term.t

type t = abstract

let rec print fmt = function
  | Pair(t, u) -> fprintf fmt "<%a, %a>" print t print u
  | Fst t -> fprintf fmt "fst(%a)" print t
  | Snd t -> fprintf fmt "snd(%a)" print t
  | Other t -> Term.print fmt t

exception Unsolvable
  
module R =
struct
  type t = unit
  type elt = abstract
  exception Inconsistent    
  let empty = ()
  let assume _ _ _ _ = ((),Literal.SetEq.empty)
  let query _ _ _ _ = assert false
end

let is_pair =
  let c = Hstring.make "%pair" in
    fun s -> Hstring.compare s c = 0

let is_fst =
  let n = Hstring.make "%fst" in
    fun s -> Hstring.compare s n = 0

let is_snd =
  let n = Hstring.make "%snd" in
    fun s -> Hstring.compare s n = 0

let rec normalize = function
  | Pair(u, v) -> 
      let u' = normalize u and v' = normalize v in
	Pair(u', v')
  | Fst p ->
      (match normalize p with
	| Pair(a, _) -> a
	| p' -> Fst p')
  | Snd p ->
      (match normalize p with
	| Pair(_, b) -> b
	| p' -> Snd p')
  | Other t -> Other t
   
let rec make t =
  match (Term.view t) with
    | {Term.f = Symbols.Name f; xs = l} ->
	begin
	  if is_pair f then
	    match l with 
	      | [lc; rc] -> Pair (make lc, make rc)
	      | _ -> assert false
	  else if is_fst f then
	    match l with
	      | [p] -> Fst (make p)
	      | _ -> assert false
	  else if is_snd f then
	    match l with
	      | [p] -> Snd (make p)
	      | _ -> assert false
	  else
	    Other t
	end
    | _ -> Other t 

let rec leaves t = 
  match (normalize t) with
    | Pair(lc, rc) -> Term.Set.union (leaves lc) (leaves rc)
    | Fst p 
    | Snd p -> leaves p
    | Other t -> Term.Set.singleton t

let compare t u =
  let rec comp t' u' =
    match t', u' with
      | Other t, Other u -> Term.compare t u
      | Other _, _ -> -1
      | _, Other _ -> 1
      | Snd p, Snd q -> comp p q
      | Snd _, _ -> -1
      | _, Snd _ -> 1
      | Fst p, Fst q -> comp p q
      | Fst _, _ -> -1
      | _, Fst _ -> 1
      | Pair(t, u), Pair(t', u') ->
	  let c = comp t t' in
	    if c = 0 then comp u u'
	    else c
  in
    comp (normalize t) (normalize u)


(* This module is a naive union-find structure for *)
(* elements of type t, implemented as a set of lists. *)
module UF = 
struct 

  module S = Set.Make  (struct
    type t = abstract list
    let rec comp l1 l2 = match l1, l2 with
      | [], [] -> 0
      | _, [] -> 1
      | [], _ -> -1
      | a::q1, b::q2 ->
	  let c = Pervasives.compare a b in
	    if c = 0 then comp q1 q2 else c
    let compare = comp
  end)
    
  type t = S.t

  let empty = S.empty
  let mem t =
    S.exists (List.mem t)

  let add t uf =
    if mem t uf then uf 
    else S.add [t] uf
    
  let classof uf t = 
    match S.elements (S.filter (List.mem t) uf) with
      | [] -> raise Not_found
      | [c] -> c
      | _ -> assert false 
	  (* on ne doit pas trouver deux classes pour un élément donné *)

  let find uf t = 
    List.hd (classof uf t)
      
  let union uf t1 t2 =
    let c1 = classof uf t1 and c2 = classof uf t2 in
    S.add (c1@c2) (S.remove c2 (S.remove c1 uf))
      
end

module Map = Map.Make (struct type t = abstract let compare = compare end)
module Set = Set.Make (struct type t = abstract let compare = compare end)

let rec subst_rec p v = function
  | Other t ->
      if Term.compare p t = 0 then v
      else Other t
  | Fst t -> Fst (subst_rec p v t)
  | Snd t -> Snd (subst_rec p v t)
  | Pair(t, u) -> Pair (subst_rec p v t, subst_rec p v u)

let subst p v r =
(*   subst_rec p v r *)
  normalize (subst_rec p v r)
      
let is_mine_t t = 
  match (make t) with 
    | Pair(_,_)
    | Fst _
    | Snd _ -> true
    | Other t -> false
	
let is_mine_r r = 
(*   if debug_pairs then  *)
(*     printf "Is pair theory %a ? " print r; *)
  match normalize r with
(*     | Fst (Other t) *)
(*     | Snd (Other t) *)
    | Other t -> 
	(* 	if debug_pairs then printf "No %a @." Term.print t; *)
	Sig.No t
    | Pair(_,_)
    | Fst _
    | Snd _ -> 
(* 	if debug_pairs then printf "Yes @."; *)
	Sig.Yes
      
let is_mine_a _ = false

let normal_form a = a

(** [decompose uf t] adds [t] and all its subterms to [uf] *)
let rec decompose uf = function
  | Pair(t, u) as p -> 
      UF.add p (decompose (decompose uf t) u)
  | Fst(t) | Snd(t) as p ->
      UF.add p (decompose uf t)
  | p -> UF.add p uf

(** [find_new_eqs uf] tries to find new equalities between elements
    of [uf] by applying simple rules.
    When no equality is found, raises Not_found. *)
let find_new_eqs uf =
  (* Three possible origins of new equalities : *)
  (* _ fst(pair(x,y)) = x *)
  let find_fst_pair uf =
    match UF.S.fold
      (fun c acc -> List.fold_left 
	(fun acc t -> match t with 
	  | Fst(e) -> 
	      (List.fold_left 
		  (fun acc e' -> match e' with 
		      Pair(a, _) when not (List.mem a c) -> (t,a)::acc
		    | _ -> acc)
		  acc (UF.classof uf e))
	  | _ -> acc) acc c)
      uf [] with
	| [] -> raise Not_found
	| lt -> lt in
  (* _ snd(pair(x,y)) = y *)
  let find_snd_pair uf =
    match UF.S.fold
      (fun c acc -> List.fold_left 
	(fun acc t -> match t with 
	  | Snd(e) -> 
	      (List.fold_left 
		  (fun acc e' -> match e' with 
		      Pair(_, b) when not (List.mem b c) -> (t,b)::acc
		    | _ -> acc)
		  acc (UF.classof uf e))
	  | _ -> acc) acc c)
      uf [] with
	| [] -> raise Not_found
	| lt -> lt in
  (* _ cons(x,y) = cons(x',y') -> x=x' /\ y=y' *)
  let find_pair_pair uf =
    let rec apparier acc = function
      | [] -> acc
      | Pair(t1,t2)::q -> 
	  let aq = List.fold_left (fun acc t ->
	    match t with Pair(u1, u2) -> 
	      let add = ref [] in
		if not (List.mem u1 (UF.classof uf t1)) then
		  add := (t1, u1)::!add;
		if not (List.mem u2 (UF.classof uf t2)) then 
		  add := (t2, u2)::!add;
		(!add)@acc
	      | _ -> assert false) [] q
	  in apparier (aq@acc) q
      | _ -> assert false in
      match UF.S.fold
	(fun c acc -> 
	  apparier acc 
	    (List.filter (function Pair(_,_) -> true | _ -> false) c))
	uf [] with
      	  | [] -> raise Not_found
	  | lt -> lt
  in
    try find_fst_pair uf
    with Not_found ->
      (try find_snd_pair uf
	with Not_found ->
	  find_pair_pair uf)
	
let solve t1 t2 = 
  if debug_pairs then 
    printf "Solving %a = %a @." print t1 print t2;
  (*   eprintf "[pair] solve %a %a@." print t1 print t2; *)
  if compare t1 t2 = 0 then [] else
    (* populate the uf *)
  let uf = decompose (decompose UF.empty t1) t2 in
    (* merge t1 and t2 *)
  let uf' = UF.union uf t1 t2 in
    (* close the uf wrt the new equalities *)
  let rec closure uf =
    try let lt = find_new_eqs uf in
	  closure 
	    (List.fold_left (fun acc (t1, t2) -> UF.union acc t1 t2) uf lt)
    with Not_found -> uf in
  let ufsat = closure uf' in
    (* For each leaf, we find a term in its class which was not
       already equal to itself *)
  let res =
    Term.Set.fold
      (fun t acc -> 
	 eprintf "@.la classe de %a est :" Term.print t;
	match (List.filter (fun e -> 
			      eprintf "%a@. " print e;
			      compare (Other t) e <> 0)
		  (UF.classof ufsat (Other t))) with
	    [] -> acc
	  | e::_ -> (t, e)::acc)
      (leaves t1)
      [] in
    if debug_pairs then 
      (printf "Results :@ ";
       List.iter (fun (t, r) -> printf "%a = %a@ ;@ " Term.print t print r) res);
    res
