open Smt_ast
open Why_ptree
open Lexing

let dpos = {pos_fname="";pos_lnum=0;pos_bol=0;pos_cnum=0}
let dloc = dpos, dpos

exception Not_Implemented

let sort_to_type pos = function
  | "Int" -> PPTint
  | "Real" -> PPTreal
  | s -> PPTexternal ([], s, pos)
  
let curry_sort_to_type (pos,sort) =
  sort_to_type pos sort

let pexp pos f = {pp_loc = pos; pp_desc = f}
let mlexpr f = pexp dloc f

let make_forall f bl  =
  let f = List.fold_left
           (fun f x -> pexp f.pp_loc
                        (PPforall ([x.var], 
                         sort_to_type f.pp_loc x.sort, [], f ))) f bl
  in f.pp_desc

let make_exists f bl =
  let f = 
    List.fold_left 
      (fun f x -> pexp f.pp_loc 
        (PPexists (x.var, sort_to_type f.pp_loc x.sort, f))) f bl
  in f.pp_desc

let make_infix op = function
  | x::l -> let f = List.fold_left 
                     (fun f x -> pexp x.pp_loc (PPinfix (f,op,x)))
                     x l
            in
              f.pp_desc
  | _ -> assert false

let make_distinct_aux elem l =
  List.map ( fun x -> PPinfix (elem, PPneq, x)) l

let make_distinct l =
  let rec aux = function
    | [] -> []
    | (x::xs) -> 
        List.rev_append (make_distinct_aux x xs) (aux xs)
  in
    List.map mlexpr (aux l)

let wequal x y = PPinfix (x, PPeq, y)

let rec make_equal = function
  | [] -> PPtrue
  | [x] -> PPtrue
  | [x;y] -> wequal x y
  | (x::y::z::xs) -> 
      PPinfix (mlexpr (wequal x y),
               PPand,
               mlexpr (make_equal (y::z::xs)))

let infix = function
  | "<" -> PPlt
  | "<=" -> PPle
  | ">" -> PPgt
  | ">=" -> PPge
  | "+" -> PPadd
  | "-" -> PPsub
  | "*" -> PPmul
  | "/" -> PPdiv
  | _ -> assert false

let rec term_to_desc = function
  | Num str -> PPconst (ConstInt str)
  | Var str -> PPvar str
  | Fun ("+" | "-" | "*" | "/" as s, [a;b]) -> 
      PPinfix (term a, infix s, term b)
  | Fun ("~", [a]) -> 
      PPprefix(PPneg, term a)
  | Fun (str,tlist) -> PPapp (str, List.map term tlist)
  | Ite (f,t1,t2) -> raise Not_Implemented (*PPapp ("ite", List.map term [t1;t2])*)

and term a = pexp a.tloc (term_to_desc a.term)

let rec form_to_desc = function
  | Flet (ident,f1,f2) -> 
      raise Not_Implemented
  | Let (ident,t,f1) ->
      raise Not_Implemented
  | Forall (binders,f1) -> 
      make_forall (formula f1) binders 
  | Exists (binders,f1) -> 
      make_exists (formula f1) binders
  | And l -> 
      make_infix PPand (List.map formula l)
  | Or l -> 
      make_infix PPor (List.map formula l)
  | Not f -> 
      PPprefix (PPnot, formula f)
  | Implies (f1,f2) -> 
      make_infix PPimplies (List.map formula [f1;f2])
  | Xor l -> 
      raise Not_Implemented
  | Iff l -> 
      make_infix PPiff (List.map formula l )
  | Fite (f,f_if,f_else) ->
      raise Not_Implemented
  | True -> PPtrue
  | False -> PPfalse
  | Fvar str -> PPvar str
  | Distinct l -> make_infix PPand (make_distinct (List.map term l))
  | Equals l -> make_equal (List.map term l)
  | Pred (("<" | "<=" | ">" | ">=" as s), [a; b]) -> 
      PPinfix (term a, infix s, term b)
  | Pred (str,tlist) -> 
      PPapp (str, List.map term tlist)

and formula f = pexp f.floc (form_to_desc f.formula)
      
let psig_to_pdef (pos,psig) =
  Logic (pos,false,[psig.pname],
	 PPredicate (List.map curry_sort_to_type psig.pargs))

let fsig_to_fdef (pos,fsig) =
  Logic (pos,false,[fsig.fname],
	 PFunction ((List.map curry_sort_to_type fsig.fargs),
                    curry_sort_to_type fsig.fres ))

let sortdecl_to_type (pos,s) =
  TypeDecl (pos,false,[], s)
  

let bench_to_why = function
  | Pblogic _ -> [] (* this seems to be comments only, throw away *)
  | Pbstatus s -> []
  | Pbextr_preds plist -> 
      List.map psig_to_pdef plist
  | Pbextr_funs flist -> 
      List.map fsig_to_fdef flist
  | Pbextr_sorts sorts -> 
      List.map sortdecl_to_type sorts
  | Pbformula (pos,f) -> 
      [Goal (pos, "", formula f)]
  | Pbassumption (pos,f) -> 
      [Axiom (pos, "", formula f)]
  | Pannotation  -> []

